/*
 * FILE --- tank_sim.hh
 *
 * PURPOSE --- To provide an example use of the class for 4'th
 * order Runge-Kutta simulation, RK4_SIM, and the class for implementing
 * n-input m-output adaptibe Radial Basis Function (RBF) neural network (NN).
 *
 * Implements the member functions of a class for performing a simulation of
 * a simple surge tank level dynamics controlled by a RBF adaptive controller.
 *
 */

// Prevent Multiple Inclusion
#ifndef TANK_SIM_HH
#define TANK_SIM_HH

// Include Files
#include <math.h>
#include "rk4_sim.hh"

class TANK_SIM : public RK4_SIM
{
public:
  TANK_SIM (double samp_time = 0.005, int _simulation_per_cycle = 10,
	    double _initial_time = 0.0, double *_initial_state = NULL);
  ~TANK_SIM ();
  
  
  double sampling_period;
  int simulation_per_cycle; // # of RK4 simulations per sampling interval
  
  // Overload DERIV function of RK4_SIM for simulating vehicle dynamics
  inline void DERIV(const double t, const double *x, 
		    const double *u, double *xdot);
  
  // the Simulation routine
  void simulate (double _input);
  
private:
  // Plant parameters
  double a;  // Cross sectional area parameter
  double b;  // Cross sectional area parameter
  double c;  // Cross sectional area of the output pipe
  double g;  // Acceleration due to graviry
};

#endif // TANK_SIM_HH
