/*
 * FILE --- tanker_sim.hh
 *
 * PURPOSE --- To provide an example use of the class for 4'th
 * order Runge-Kutta simulation, RK4_SIM, and the class for implementing
 * n-input m-output Multilayer Perceptron (MLP) neural network (NN).
 *
 * Implements the member functions of a class for performing a simulation of
 * a simple cargo ship whose steering is controlled by a MLP controller.
 *
 */

// Prevent Multiple Inclusion
#ifndef TANKER_SIM_HH
#define TANKER_SIM_HH

// Include Files
#include <math.h>
#include "rk4_sim.hh"

class TANKER_SIM : public RK4_SIM
{
public:
  TANKER_SIM (double samp_time = 0.005, int _simulation_per_cycle = 10,
	      double _initial_time = 0.0, double *_initial_state = NULL);
  ~TANKER_SIM ();
  
  
  double sampling_period;
  int simulation_per_cycle; // # of RK4 simulations per sampling interval
  
  // Overload DERIV function of RK4_SIM for simulating vehicle dynamics
  inline void DERIV(const double t, const double *x, 
		    const double *u, double *xdot);
  
  // the Simulation routine
  void simulate (double _input);

private:
  // Plant parameters
  double lenght;  // Length of the ship (in meters)
  double speed;   // Nominal speed (in meters/sec)
  double abar;  // Parameters for nonlinearity
  double bbar;

  double K_0;
  double tau_10;
  double tau_20;
  double tau_30;

  double K;
  double tau_1;
  double tau_2;
  double tau_3;
};

#endif // TANKER_SIM_HH
