/* 
 * FILE --- matrix.cc
 *
 * PURPOSE --- Source file for a class for implementing a matrix of doubles
 * together with matrix operations such as multiplication, transpose, inverse,
 * etc. It is intended to be used with the control library.
 */

// Include Files
#include <math.h>
#include "matrix.hh"

// Define this macro if you want to use this utility without linking
// to the RCS Library
#ifdef NO_RCS
#include <stdio.h>            /* printf function */
#else
#include "rcs_prnt.hh"          /* rcs_print_ functions */
#endif

/*************************************************************/

MATRIX::MATRIX(int _row = 2, int _col = 2, double _val = 0.0)
{
  // No error at start
  matrix_err = MATRIX_NO_ERROR;
  row = _row;
  col = _col;
  if (row < 1 || col < 1)
    {
      matrix_err = INVALID_DIMENTION;
#ifdef NO_RCS
      printf("MATRIX: ERROR --- %d!!!\n", matrix_err);
      printf("Invalid dimension. Cant create matrix ...\n");
#else
      rcs_print("MATRIX: ERROR --- %d!!!\n", matrix_err);
      rcs_print("Invalid dimension. Cant create matrix ...\n");
#endif
    }
  else
    {
      allocate_memory();
    }
  
  // Set the elements
  set_matrix(_val);
}

MATRIX::MATRIX(int _row, int _col, double **_elements)
{
  // No error at start
  matrix_err = MATRIX_NO_ERROR;
  row = _row;
  col = _col;
  if (row < 1 || col < 1)
    {
      matrix_err = INVALID_DIMENTION;
#ifdef NO_RCS
      printf("MATRIX: ERROR --- %d!!!\n", matrix_err);
      printf("Invalid dimension. Cant create matrix ...\n");
#else
      rcs_print("MATRIX: ERROR --- %d!!!\n", matrix_err);
      rcs_print("Invalid dimension. Cant create matrix ...\n");
#endif
    }
  else
    {
      allocate_memory();
    }
  
  // Set the elements
  if (_elements == NULL)
    {
      set_matrix(0.0);
    }
  else
    {
      set_matrix((const double **)_elements);
    }
}

MATRIX::MATRIX(const MATRIX &_matrix)
{
  // No error at start
  matrix_err = MATRIX_NO_ERROR;
  row = _matrix.row;
  col = _matrix.col;
  allocate_memory();
  
  // Set the elements
  set_matrix((const double **)_matrix.elements);
}

MATRIX::~MATRIX()
{
  free_memory();
}

void MATRIX::free_memory(void)
{
  for (int ii = 0; ii < row; ii++)
    {
      delete [] elements[ii];
    }
  delete [] elements;
}

void MATRIX::allocate_memory(void)
{
  if ((elements = new double*[row]) == NULL)
    {
      matrix_err = MEMORY_ERROR;
#ifdef NO_RCS
      printf("MATRIX: ERROR --- %d!!!\n", matrix_err);
      printf("Not Enought Memory. Cant create matrix ...\n");
#else
      rcs_print("MATRIX: ERROR --- %d!!!\n", matrix_err);
      rcs_print("Not Enought Memory. Cant create matrix ...\n");
#endif
    }
  for (int ii = 0; ii < row; ii++)
    {
      if ((elements[ii] = new double[col]) == NULL)
	{
	  matrix_err = MEMORY_ERROR;
#ifdef NO_RCS
	  printf("MATRIX: ERROR --- %d!!!\n", matrix_err);
	  printf("Not Enought Memory. Cant create matrix ...\n");
#else
	  rcs_print("MATRIX: ERROR --- %d!!!\n", matrix_err);
	  rcs_print("Not Enought Memory. Cant create matrix ...\n");
#endif
	}
    }
}

void MATRIX::set_matrix(double _val)
{
  int ii, jj; // Index variales for the loops
  
  for (ii = 0; ii < row; ii++)
    {
      for (jj = 0; jj < col; jj++)
	{
	  elements[ii][jj] = _val;
	}
    }
}

void MATRIX::set_matrix(const double **_elements)
{
  int ii, jj; // Index variales for the loops
  
  for (ii = 0; ii < row; ii++)
    {
      for (jj = 0; jj < col; jj++)
	{
	  elements[ii][jj] = _elements[ii][jj];
	}
    }
}

void MATRIX::set_matrix(const double *_elements)
{
  if (col == 1)
    {
      set_col(0, _elements);
    }
  else if (row == 1)
    {
      set_row(0, _elements);
    }
  else
    {
      matrix_err = INVALID_SET_MATRIX;
#ifdef NO_RCS
      printf("MATRIX:set_matrix: ERROR --- %d!!!\n", matrix_err);
      printf("Matrix not vector. Provide two dimetional array ...\n");
#else
      rcs_print("MATRIX:set_matrix: ERROR --- %d!!!\n", matrix_err);
      rcs_print("Matrix not vector. Provide two dimetional array ...\n");
#endif

    }
}


void MATRIX::set_diagonal(double _val)
{
  if (row != col)
    {
      // Find the smaller of # of rows or columns
      //      int s = (row < col) ? row : col;
      //      for (int ii = 0; ii < s; ii++)
      //	{
      //	  elements[ii][ii] = _val;
      //	}
      matrix_err = NONSQUARE_MATRIX;
#ifdef NO_RCS
      printf("MATRIX:set_diagonal: ERROR --- %d!!!\n", matrix_err);
      printf("Nonsqure matrix. Cant set the diagonal ...\n");
#else
      rcs_print("MATRIX:set_diagonal: ERROR --- %d!!!\n", matrix_err);
      rcs_print("Nonsqure matrix. Cant set the diagonal ...\n");
#endif
    }
  else
    {
      for (int ii = 0; ii < row; ii++)
	{
	  elements[ii][ii] = _val;
	}
    }
}

void MATRIX::set_diagonal(const double *_elements)
{
  if (row != col)
    {
      // Find the smaller of # of rows or columns
      //  int s = (row < col) ? row : col;
      //  for (int ii = 0; ii < s; ii++)
      //    {
      //      elements[ii][ii] = _elements[ii];
      //    }
      matrix_err = NONSQUARE_MATRIX;
#ifdef NO_RCS
      printf("MATRIX:set_diagonal: ERROR --- %d!!!\n", matrix_err);
      printf("Nonsqure matrix. Cant set the diagonal ...\n");
#else
      rcs_print("MATRIX:set_diagonal: ERROR --- %d!!!\n", matrix_err);
      rcs_print("Nonsqure matrix. Cant set the diagonal ...\n");
#endif
    }
  else
    {
      for (int ii = 0; ii < row; ii++)
	{
	  elements[ii][ii] = _elements[ii];
	}
    }
}

void MATRIX::set_row(int _row, double _val)
{
  if (_row < 0 || _row >= row)
    {
      matrix_err = ILLEGAL_INDEX;
#ifdef NO_RCS
      printf("MATRIX:set_row: ERROR --- %d!!!\n", matrix_err);
      printf("Invalid row index. Cant set row ...\n");
#else
      rcs_print("MATRIX:set_row: ERROR --- %d!!!\n", matrix_err);
      rcs_print("Invalid row index. Cant set row  ...\n");
#endif
    }
  else
    {
      for (int ii = 0; ii < col; ii++)
	{
	  elements[_row][ii] = _val;
	}
    }
}

void MATRIX::set_row(int _row, const double *_elements)
{
  if (_row < 0 || _row >= row)
    {
      matrix_err = ILLEGAL_INDEX;
#ifdef NO_RCS
      printf("MATRIX:set_row: ERROR --- %d!!!\n", matrix_err);
      printf("Invalid row index. Cant set row ...\n");
#else
      rcs_print("MATRIX:set_row: ERROR --- %d!!!\n", matrix_err);
      rcs_print("Invalid row index. Cant set row  ...\n");
#endif
    }
  else
    {
      for (int ii = 0; ii < col; ii++)
	{
	  elements[_row][ii] = _elements[ii];
	}
    }
}

void MATRIX::set_col(int _col, double _val)
{
  if (_col < 0 || _col >= col)
    {
      matrix_err = ILLEGAL_INDEX;
#ifdef NO_RCS
      printf("MATRIX:set_col: ERROR --- %d!!!\n", matrix_err);
      printf("Invalid column index. Cant set column ...\n");
#else
      rcs_print("MATRIX:set_col: ERROR --- %d!!!\n", matrix_err);
      rcs_print("Invalid column index. Cant set column  ...\n");
#endif
    }
  else
    {
      for (int ii = 0; ii < row; ii++)
	{
	  elements[ii][_col] = _val;
	}
    }
}

void MATRIX::set_col(int _col, const double *_elements)
{
  if (_col < 0 || _col >= col)
    {
      matrix_err = ILLEGAL_INDEX;
#ifdef NO_RCS
      printf("MATRIX:set_col: ERROR --- %d!!!\n", matrix_err);
      printf("Invalid column index. Cant set column ...\n");
#else
      rcs_print("MATRIX:set_col: ERROR --- %d!!!\n", matrix_err);
      rcs_print("Invalid column index. Cant set column  ...\n");
#endif
    }
  else
    {
      for (int ii = 0; ii < row; ii++)
	{
	  elements[ii][_col] = _elements[ii];
	}
    }
}

double MATRIX::get_element(int _row, int _col) const
{
  if(_row < 0 || _row >= row || _col < 0 || _col >= col)
    {
#ifdef NO_RCS
      printf("MATRIX:get_element: ERROR !!!\n");
      printf("Invalid index. Zero returned ...\n");
#else
      rcs_print("MATRIX:get_element: ERROR !!!\n");
      rcs_print("Invalid index. Zero returned ...\n");
#endif
      return (0.0);
    }
  else
    {
      return (elements[_row][_col]);
    }
}


void MATRIX::set_element(int _row, int _col, double _val)
{
  if (_row < 0 || _row >= row || _col < 0 || _col >= col)
    {
      matrix_err = ILLEGAL_INDEX;
#ifdef NO_RCS
      printf("MATRIX:set_element: ERROR --- %d!!!\n", matrix_err);
      printf("Invalid index. Cant set element ...\n");
#else
      rcs_print("MATRIX:set_element: ERROR --- %d!!!\n", matrix_err);
      rcs_print("Invalid index. Cant set element ...\n");
#endif
    }
  else
    {
      elements[_row][_col] = _val;
    }
}

void MATRIX::add_to_element(int _row, int _col, double _val)
{
  if (_row < 0 || _row >= row || _col < 0 || _col >= col)
    {
      matrix_err = ILLEGAL_INDEX;
#ifdef NO_RCS
      printf("MATRIX:add_to_element: ERROR --- %d!!!\n", matrix_err);
      printf("Invalid index. Cant add to element ...\n");
#else
      rcs_print("MATRIX:add_to_element: ERROR --- %d!!!\n", matrix_err);
      rcs_print("Invalid index. Cant add to element ...\n");
#endif
    }
  else
    {
      elements[_row][_col] += _val;
    }
}

void MATRIX::subtract_from_element(int _row, int _col, double _val)
{
  if (_row < 0 || _row >= row || _col < 0 || _col >= col)
    {
      matrix_err = ILLEGAL_INDEX;
#ifdef NO_RCS
      printf("MATRIX:subtract_from_element: ERROR --- %d!!!\n", matrix_err);
      printf("Invalid index. Cant subtract from element ...\n");
#else
      rcs_print("MATRIX:subtract_from_element: ERROR --- %d!!!\n", matrix_err);
      rcs_print("Invalid index. Cant subtract from element ...\n");
#endif
    }
  else
    {
      elements[_row][_col] -= _val;
    }
}

void MATRIX::multiply_element(int _row, int _col, double _val)
{
  if (_row < 0 || _row >= row || _col < 0 || _col >= col)
    {
      matrix_err = ILLEGAL_INDEX;
#ifdef NO_RCS
      printf("MATRIX:multiply_element: ERROR --- %d!!!\n", matrix_err);
      printf("Invalid index. Cant multiply element ...\n");
#else
      rcs_print("MATRIX:multiply_element: ERROR --- %d!!!\n", matrix_err);
      rcs_print("Invalid index. Cant multiply element ...\n");
#endif
    }
  else
    {
      elements[_row][_col] *= _val;
    }
}

void MATRIX::divide_element(int _row, int _col, double _val)
{
  if (_row < 0 || _row >= row || _col < 0 || _col >= col)
    {
      matrix_err = ILLEGAL_INDEX;
#ifdef NO_RCS
      printf("MATRIX:divide_element: ERROR --- %d!!!\n", matrix_err);
      printf("Invalid index. Cant divide element ...\n");
#else
      rcs_print("MATRIX:divide_element: ERROR --- %d!!!\n", matrix_err);
      rcs_print("Invalid index. Cant divide element ...\n");
#endif
    }
  else
    {
      if (_val)
	{
	  elements[_row][_col] /= _val;
	}
      else
	{
	  matrix_err = DIVIDE_BY_ZERO;
#ifdef NO_RCS
	  printf("MATRIX:divide_element: ERROR --- %d!!!\n", matrix_err);
	  printf("Divide by zero. Cant divide element ...\n");
#else
	  rcs_print("MATRIX:divide_element: ERROR --- %d!!!\n", matrix_err);
	  rcs_print("Divide by zero. Cant divide element ...\n");
#endif
	}
    }
}

void MATRIX::add_to_row(int _row, double _val)
{
  if (_row < 0 || _row >= row)
    {
      matrix_err = ILLEGAL_INDEX;
#ifdef NO_RCS
      printf("MATRIX:add_to_row: ERROR --- %d!!!\n", matrix_err);
      printf("Invalid row index. Cant add to row ...\n");
#else
      rcs_print("MATRIX:add_to_row: ERROR --- %d!!!\n", matrix_err);
      rcs_print("Invalid row index. Cant add to row  ...\n");
#endif
    }
  else
    {
      for (int ii = 0; ii < col; ii++)
	{
	  elements[_row][ii] += _val;
	}
    }
}

void MATRIX::subtract_from_row(int _row, double _val)
{
  if (_row < 0 || _row >= row)
    {
      matrix_err = ILLEGAL_INDEX;
#ifdef NO_RCS
      printf("MATRIX:subtract_from_row: ERROR --- %d!!!\n", matrix_err);
      printf("Invalid row index. Cant subtract from row ...\n");
#else
      rcs_print("MATRIX:subtract_from_row: ERROR --- %d!!!\n", matrix_err);
      rcs_print("Invalid row index. Cant subtract from row  ...\n");
#endif
    }
  else
    {
      for (int ii = 0; ii < col; ii++)
	{
	  elements[_row][ii] -= _val;
	}
    }
}

void MATRIX::multiply_row(int _row, double _val)
{
  if (_row < 0 || _row >= row)
    {
      matrix_err = ILLEGAL_INDEX;
#ifdef NO_RCS
      printf("MATRIX:multiply_row: ERROR --- %d!!!\n", matrix_err);
      printf("Invalid row index. Cant multiply row ...\n");
#else
      rcs_print("MATRIX:multiply_row: ERROR --- %d!!!\n", matrix_err);
      rcs_print("Invalid row index. Cant multiply row  ...\n");
#endif
    }
  else
    {
      for (int ii = 0; ii < col; ii++)
	{
	  elements[_row][ii] *= _val;
	}
    }
}

void MATRIX::devide_row(int _row, double _val)
{
  if (_row < 0 || _row >= row)
    {
      matrix_err = ILLEGAL_INDEX;
#ifdef NO_RCS
      printf("MATRIX:divide_row: ERROR --- %d!!!\n", matrix_err);
      printf("Invalid row index. Cant divide row ...\n");
#else
      rcs_print("MATRIX:divide_row: ERROR --- %d!!!\n", matrix_err);
      rcs_print("Invalid row index. Cant divide row  ...\n");
#endif
    }
  else
    {
      if (_val)
	{
	  for (int ii = 0; ii < col; ii++)
	    {
	      elements[_row][ii] /= _val;
	    }
	}
      else
	{
	  matrix_err = DIVIDE_BY_ZERO;
#ifdef NO_RCS
	  printf("MATRIX:divide_row: ERROR --- %d!!!\n", matrix_err);
	  printf("Divide by zero. Cant divide row ...\n");
#else
	  rcs_print("MATRIX:divide_row: ERROR --- %d!!!\n", matrix_err);
	  rcs_print("Divide by zero. Cant divide row ...\n");
#endif
	}
    }
}

void MATRIX::add_to_col(int _col, double _val)
{
  if (_col < 0 || _col >= col)
    {
      matrix_err = ILLEGAL_INDEX;
#ifdef NO_RCS
      printf("MATRIX:add_to_col: ERROR --- %d!!!\n", matrix_err);
      printf("Invalid column index. Cant add to column ...\n");
#else
      rcs_print("MATRIX:add_to_col: ERROR --- %d!!!\n", matrix_err);
      rcs_print("Invalid column index. Cant add to column  ...\n");
#endif
    }
  else
    {
      for (int ii = 0; ii < row; ii++)
	{
	  elements[ii][_col] += _val;
	}
    }
}

void MATRIX::subtract_from_col(int _col, double _val)
{
  if (_col < 0 || _col >= col)
    {
      matrix_err = ILLEGAL_INDEX;
#ifdef NO_RCS
      printf("MATRIX:subtract_from_col: ERROR --- %d!!!\n", matrix_err);
      printf("Invalid column index. Cant subtract from column ...\n");
#else
      rcs_print("MATRIX:subtract_from_col: ERROR --- %d!!!\n", matrix_err);
      rcs_print("Invalid column index. Cant subtruct from column  ...\n");
#endif
    }
  else
    {
      for (int ii = 0; ii < row; ii++)
	{
	  elements[ii][_col] -= _val;
	}
    }
}

void MATRIX::multiply_col(int _col, double _val)
{
  if (_col < 0 || _col >= col)
    {
      matrix_err = ILLEGAL_INDEX;
#ifdef NO_RCS
      printf("MATRIX:multiply_col: ERROR --- %d!!!\n", matrix_err);
      printf("Invalid column index. Cant multiply column ...\n");
#else
      rcs_print("MATRIX:multiply_col: ERROR --- %d!!!\n", matrix_err);
      rcs_print("Invalid column index. Cant multiply column  ...\n");
#endif
    }
  else
    {
      for (int ii = 0; ii < row; ii++)
	{
	  elements[ii][_col] *= _val;
	}
    }
}

void MATRIX::devide_col(int _col, double _val)
{
  if (_col < 0 || _col >= col)
    {
      matrix_err = ILLEGAL_INDEX;
#ifdef NO_RCS
      printf("MATRIX:divide_col: ERROR --- %d!!!\n", matrix_err);
      printf("Invalid column index. Cant divide column ...\n");
#else
      rcs_print("MATRIX:divide_col: ERROR --- %d!!!\n", matrix_err);
      rcs_print("Invalid column index. Cant divide column  ...\n");
#endif
    }
  else
    {
      if (_val)
	{
	  for (int ii = 0; ii < row; ii++)
	    {
	      elements[ii][_col] *= _val;
	    }
	}
      else
	{
	  matrix_err = DIVIDE_BY_ZERO;
#ifdef NO_RCS
	  printf("MATRIX:divide_col: ERROR --- %d!!!\n", matrix_err);
	  printf("Divide by zero. Cant divide column ...\n");
#else
	  rcs_print("MATRIX:divide_col: ERROR --- %d!!!\n", matrix_err);
	  rcs_print("Divide by zero. Cant divide column ...\n");
#endif
	}
    }
}

void MATRIX::add_rows(int _source_row, int _destination_row, double gain)
{
  if ((_source_row < 0 || _source_row >= row) || 
      (_destination_row < 0 || _destination_row >= row))
    {
      matrix_err = ILLEGAL_INDEX;
#ifdef NO_RCS
      printf("MATRIX:add_rows: ERROR --- %d!!!\n", matrix_err);
      printf("Invalid index. Cant add rows ...\n");
#else
      rcs_print("MATRIX:add_rows: ERROR --- %d!!!\n", matrix_err);
      rcs_print("Invalid index. Cant add rows  ...\n");
#endif
    }
  else
    {
      for (int ii = 0; ii < col; ii++)
	{
	  elements[_destination_row][ii] += gain*elements[_source_row][ii];
	}
    }
}

void MATRIX::add_cols(int _source_col, int _destination_col, double gain)
{
  if ((_source_col < 0 || _source_col >= col) || 
      (_destination_col < 0 || _destination_col >= col))
    {
      matrix_err = ILLEGAL_INDEX;
#ifdef NO_RCS
      printf("MATRIX:add_cols: ERROR --- %d!!!\n", matrix_err);
      printf("Invalid index. Cant add columns ...\n");
#else
      rcs_print("MATRIX:add_cols: ERROR --- %d!!!\n", matrix_err);
      rcs_print("Invalid index. Cant add columns  ...\n");
#endif
    }
  else
    {
      for (int ii = 0; ii < row; ii++)
	{
	  elements[ii][_destination_col] += gain*elements[ii][_source_col];
	}
    }
}

MATRIX inv(const MATRIX &_matrix)
{
  int ii, jj; // Index variales for the loops

  // Get the # of rows and columns of the input matrix
  int r = _matrix.get_num_rows();
  int c = _matrix.get_num_cols();

  // Temporary variables
  double temp;
  MATRIX ans(r, c);
  MATRIX copy = _matrix;
  
  // Check whether square or not
  if (r != c)  
    { // Nonsquare matrix
      ans.matrix_err = NONSQUARE_MATRIX;
#ifdef NO_RCS
      printf("inv: ERROR --- Nonsqure matrix. Cant invert!!!\n");
      printf("Use pinv for the pseudoinverse ...\n");
#else
      rcs_print("inv: ERROR --- Nonsqure matrix. Cant invert!!!\n");
      rcs_print("Use pinv for the pseudoinverse ...\n");
#endif
      // Return matrix of zeros
      ans.set_matrix(0.0);
      return (ans);
    }
  else // Square
    { // Inverts using row operations. Lame, but works most times.
      
      // The answer is identity initially
      ans.set_diagonal(1.0);
      
      for (ii = 0; ii < r; ii++) 
	{
	  // Get the next diagonal element
	  temp = copy.get_element(ii, ii);

	  if (temp == 0.0)
	    {
	      ans.matrix_err = NONINVERTIBLE_MATRIX;
#ifdef NO_RCS
	      printf("inv: ERROR --- Noninvertible matrix. Cant invert!!!\n");
	      printf("Check whether your matrix is singular ...\n");
#else
	      rcs_print("inv: ERROR --- Noninvertible matrix. Cant invert!!!\n");
	      rcs_print("Check whether your matrix is singular ...\n");
#endif
	      // Return matrix of zeros
	      ans.set_matrix(0.0);
	      return (ans);
	    }
	  else 
	    {
	      ans.multiply_row(ii, 1.0/temp);
	      copy.multiply_row(ii, 1.0/temp);
	    }
	  for (jj = 0; jj < r; jj++)
	    if (ii != jj) 
	      {
		ans.add_rows(ii, jj, -copy.get_element(jj, ii));
		copy.add_rows(ii, jj, -copy.get_element(jj, ii));
	      }
	}
      return (ans);
    }
}

MATRIX pinv(const MATRIX &_matrix)
{
  // Get the # of rows and columns of the input matrix
  int r = _matrix.get_num_rows();
  int c = _matrix.get_num_cols();
  
  // Inverts using row operations. Lame, but works most times.
  if (r == c)
    {
      return(inv(_matrix));
    }
  else
    {
      if (r < c)
	{ // Return the right pseudoinverse
	  return ( transpose(_matrix) * inv(_matrix * transpose(_matrix)) );
	}        
      else
	{ // Return the left pseudoinverse
	  return ( inv(transpose(_matrix) * _matrix) * transpose(_matrix) );
	}        
    }
}

MATRIX transpose(const MATRIX &_matrix)
{
  // Get the # of rows and columns of the input matrix
  int r = _matrix.get_num_rows();
  int c = _matrix.get_num_cols();
  
  int ii, jj; // Index variales for the loops

  MATRIX ans(c,r);

  for (ii = 0; ii < r; ii++)
    {
      for (jj = 0; jj < c; jj++)
	{
	  ans.set_element(jj, ii, _matrix.get_element(ii, jj));
	}
    }
  return (ans);
}

double trace(const MATRIX &_matrix)
{
  // Get the # of rows and columns of the input matrix
  int r = _matrix.get_num_rows();
  int c = _matrix.get_num_cols();

  int ii; // Index variales for the loops
  double ans = 0.0;  // The variabe to hold the answer

  // Find the smaller of the rows and columns
  int s = (r < c ? r : c);

  for (ii = 0; ii < s; ii++)
    {
      ans += _matrix.get_element(ii, ii);
    }
  return (ans);
}

double norm(const MATRIX &_matrix)
{
  return ( sqrt(trace(_matrix * transpose(_matrix))) );
}

MATRIX &MATRIX::operator=(const MATRIX &_matrix2)
{
  int ii, jj; // Index variales for the loops

  if (&_matrix2 != this)
  {
    free_memory();
    row = _matrix2.row;
    col = _matrix2.col;
    allocate_memory();
    for (ii = 0; ii < row; ii++) 
    {
      for (jj = 0; jj < col; jj++)
      {
	elements[ii][jj] = _matrix2.elements[ii][jj];
      }
    }
  }
  return (*this);
}

const MATRIX &MATRIX::operator+=(const MATRIX &_matrix2)
{
  int ii, jj; // Index variales for the loops

  if ((row != _matrix2.row) || (col != _matrix2.col))
    {
      matrix_err = SIZE_MISMATCH;
    }
  else
    {
      for (ii = 0; ii < row; ii++)
	{
	  for (jj = 0; jj < col; jj++)
	    {
	      elements[ii][jj] += _matrix2.elements[ii][jj];
	    }
	}
    }
  return (*this);
}


const MATRIX &MATRIX::operator-=(const MATRIX &_matrix2)
{
  int ii, jj; // Index variales for the loops

  if ((row != _matrix2.row) || (col != _matrix2.col))
    {
      matrix_err = SIZE_MISMATCH;
    }
  else
    {
      for (ii = 0; ii < row; ii++)
	{
	  for (jj = 0; jj < col; jj++)
	    {
	      elements[ii][jj] -= _matrix2.elements[ii][jj];
	    }
	}
    }
  return (*this);
}

const MATRIX &MATRIX::operator*=(const double _val)
{
  int ii, jj; // Index variales for the loops

  for (ii = 0; ii < row; ii++)
    {
      for (jj = 0; jj < col; jj++)
	{
	  elements[ii][jj] = _val*elements[ii][jj];
	}
    }
  return (*this);
}

MATRIX MATRIX::operator+(const MATRIX &_matrix2) const
{
  int ii, jj; // Index variales for the loops

  MATRIX ans = *this;

  if ((row != _matrix2.row) || (col != _matrix2.col))
    {
      ans.matrix_err = SIZE_MISMATCH;
    }
  else
    {
      for (ii = 0; ii < row; ii++)
	{
	  for (jj = 0; jj < col; jj++)
	    {
	      ans.elements[ii][jj] += _matrix2.elements[ii][jj];
	    }
	}
    }
  return (ans);
}

MATRIX MATRIX::operator-(const MATRIX &_matrix2) const
{
  int ii, jj; // Index variales for the loops

  MATRIX ans = *this;

  if ((row != _matrix2.row) || (col != _matrix2.col))
    {
      ans.matrix_err = SIZE_MISMATCH;
    }
  else
    {
      for (ii = 0; ii < row; ii++)
	{
	  for (jj = 0; jj < col; jj++)
	    {
	      ans.elements[ii][jj] -= _matrix2.elements[ii][jj];
	    }
	}
    }
  return (ans);
}


MATRIX MATRIX::operator*(const MATRIX &_matrix2) const
{
  int ii, jj, kk; // Index variales for the loops

  MATRIX ans(row, _matrix2.col);

  if (col != _matrix2.row)
    {
      ans.matrix_err = SIZE_MISMATCH;
    }
  else
    {
      for (ii = 0; ii < row; ii++)
	{
	  for (jj = 0; jj < _matrix2.col; jj++)
	    {
	      for (kk = 0; kk < col; kk++)
		{
		  ans.elements[ii][jj] += 
		    elements[ii][kk] * _matrix2.elements[kk][jj];
		}
	    }
	}
    }
  return (ans);
}


MATRIX  MATRIX::operator*(const double _val)
{
  int ii, jj; // Index variales for the loops

  MATRIX ans(row,col);

  for (ii = 0; ii < row; ii++)
    {
      for (jj = 0; jj < col; jj++)
	{
	  ans.elements[ii][jj] = _val*elements[ii][jj];
	}
    }
  return (ans);
}

/************************************************************/


MATRIX operator*(const double _val, const MATRIX &_matrix)
{
  int ii, jj; // Index variales for the loops
 
  MATRIX ans(_matrix.row,_matrix.col);

  for (ii = 0; ii < _matrix.row; ii++)
    {
      for (jj = 0; jj < _matrix.col; jj++)
	{
	  ans.elements[ii][jj] = _val*_matrix.elements[ii][jj];
	}
    }
  return (ans);
}



/*********************************************************************

MATRIX inverse(const MATRIX &_matrix)
{
  int ii, jj; // Index variales for the loops
  int row = _matrix.get_row();
  int col = _matrix.get_row();
  MATRIX ans(row, row);
  MATRIX copy = _matrix;
  
  // inverts using row operations. Lame, but works most times.
  if (row != col)
    {
      ans.matrix_err = NONSQUARE_MATRIX;
    }
  else
    {
      ans.set_diagonal(1.0);
      for (ii = 0; ii < row; ii++) 
	{
	  if (copy.elements[ii][ii] == 0.0)
	    {
	      ans.matrix_err = NONINVERTIBLE_MATRIX;
	    }
	  else 
	    {
	      ans.multiply_row(ii, 1.0/copy.elements[ii][ii]);
	      copy.multiply_row(ii, 1.0/copy.elements[ii][ii]);
	    }
	  for (jj = 0; jj < row; jj++)
	    if (ii != jj) 
	      {
		ans.add_rows(ii, jj, -copy.elements[jj][ii]);
		copy.add_rows(ii, jj, -copy.elements[jj][ii]);
	      }
	}
    }
  return (ans);
}


MATRIX transpose(const MATRIX &_matrix)
{
  int ii, jj; // Index variales for the loops
  int row = _matrix.get_row();
  int col = _matrix.get_row();
  MATRIX ans(col,row);

  for (ii = 0; ii < row; ii++)
    {
      for (jj = 0; jj < col; jj++)
	{
	  ans.elements[jj][ii] = _matrix.elements[ii][jj];
	}
    }
  return (ans);
}

double trace(const MATRIX &_matrix)
{
  int ii; // Index variales for the loops
  int row = _matrix.get_row();
  int col = _matrix.get_row();
  double ans = 0.0;

  int dim = (row < col ? row : col);

  for (ii = 0; ii < dim; ii++)
    {
      ans += _matrix.elements[ii][ii];
    }
  return (ans);
}


MATRIX operator*(const double _val, const MATRIX &_matrix)
{
  int ii, jj; // Index variales for the loops
  int row = _matrix.get_row();
  int col = _matrix.get_row();
  MATRIX ans(row,col);

  for (ii = 0; ii < row; ii++)
    {
      for (jj = 0; jj < col; jj++)
	{
	  ans.elements[ii][jj] = _val*_matrix.elements[ii][jj];
	}
    }
  return (ans);
}

MATRIX operator*(const MATRIX &_matrix, const double _val)
{
  int ii, jj; // Index variales for the loops
  int row = _matrix.get_row();
  int col = _matrix.get_row();
  MATRIX ans(row,col);

  for (ii = 0; ii < row; ii++)
    {
      for (jj = 0; jj < col; jj++)
	{
	  ans.elements[ii][jj] = _val*_matrix.elements[ii][jj];
	}
    }
  return (ans);
}

**********************************************************************/
